/***************************************************************************
 *   Copyright (C) 2005 by Alejandro Vidal Mata & Javier Vidal Mata        *
 *                                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 *   The full GNU General Public License is included in this distribution  *
 *   in the file called COPYING.                                           *
 *                                                                         *
 *   This project haven't a warranty. The project developers haven't       *
 *   responsible for whatever happens from using.                          *
 *                                                                         *
 *   Authors:                                                              *
 *    Alejandro Vidal Mata <alex14@gmail.com>                              *
 *    Javier Vidal Mata <javitu@gmail.com>                                 *
 *                                                                         *
 *   Based on:                                                             *
 *    pbe5.c by Pedro Ramalhais <pmr09313@students.fct.unl.pt>             *
 ***************************************************************************/

#include <linux/compiler.h>
#include <linux/version.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
#include <linux/config.h>
#else
#include <linux/autoconf.h>
#endif

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/netdevice.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <asm/io.h>

#define DRV_NAME		"fsam7440"
#define DRV_VERSION		"1.0"
#define DRV_DESCRIPTION		"SW RF kill switch for Fujitsu-Siemens AMILO M 7440"
#define DRV_AUTHOR		"Alejandro Vidal Mata & Javier Vidal Mata"
#define DRV_LICENSE		"GPL"

static int radio = 1;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)

MODULE_PARM(radio, "i");

#else /* LINUX_VERSION_CODE < 2.6.0 */

#include <linux/moduleparam.h>
module_param(radio, int, 1);

#endif /* LINUX_VERSION_CODE < 2.6.0 */

MODULE_PARM_DESC(radio, "Controls state of radio when the module is loaded (1=on, 0=off)");

MODULE_DESCRIPTION(DRV_DESCRIPTION);
MODULE_AUTHOR(DRV_AUTHOR);
MODULE_LICENSE(DRV_LICENSE);

/*
 * These values were obtained from disassembling the PM.exe program
 * installed in the Fujitsu-Siemens AMILO M 7440
 */
#define AMILOM7440_PORT1	0x118f
#define AMILOM7440_PORT2	0x118e

#define AMILOM7440_RADIO_ON1	0x12
#define AMILOM7440_RADIO_ON2	0x80
#define AMILOM7440_RADIO_OFF1	0x10
#define AMILOM7440_RADIO_OFF2	0x00

// Some "booleans" =;-)
#define AMILOM7440_RADIO_OFF	0
#define AMILOM7440_RADIO_ON	1

static int amilom7440_radio_status = AMILOM7440_RADIO_OFF;

unsigned char amilom7440_get_radio(void)
{
	unsigned char val1 = 0x00;
	unsigned char val2 = 0x00;
	
	val1 = inb(AMILOM7440_PORT1); // Read the values of this ports to know the
	val2 = inb(AMILOM7440_PORT2); // state of wireless card.

	if ((val1 == AMILOM7440_RADIO_ON1) && (val2 == AMILOM7440_RADIO_ON2)) // Return the state
		return AMILOM7440_RADIO_ON;
	else
        	return AMILOM7440_RADIO_OFF;
}

static void amilom7440_set_radio(int state_set)
{
	amilom7440_radio_status = amilom7440_get_radio(); // Get the state of wireless card.

	if (amilom7440_radio_status != state_set) { // If it need changes
        	if (state_set == AMILOM7440_RADIO_ON) { // Turn ON/OFF
			outb(AMILOM7440_RADIO_ON1, AMILOM7440_PORT1);
			outb(AMILOM7440_RADIO_ON2, AMILOM7440_PORT2);
        	} else {
			outb(AMILOM7440_RADIO_OFF1, AMILOM7440_PORT1);
			outb(AMILOM7440_RADIO_OFF2, AMILOM7440_PORT2);
        	}

		amilom7440_radio_status = amilom7440_get_radio(); // Get the state of wireless card again
		
		if (state_set == amilom7440_get_radio()) { // Check if the state has changed correctly
			printk(KERN_INFO DRV_NAME ": Radio turned %s\n", // Notify if it has changed correctly or not
			   (state_set  == AMILOM7440_RADIO_ON) ? "ON" : "OFF");
		} else {
			printk(KERN_INFO DRV_NAME ": ERROR: Radio don't turned %s\n",
			   (state_set  == AMILOM7440_RADIO_ON) ? "ON" : "OFF");
		}
	} else {
		printk(KERN_INFO DRV_NAME ": Radio already %s\n", // It didn't need changes
		  (state_set  == AMILOM7440_RADIO_ON) ? "ON" : "OFF");
	}
}


/*
 * Proc Stuff
 */
static struct proc_dir_entry *dir_base = NULL;

static int proc_set_radio(struct file *file, const char *buffer, 
			  unsigned long count, void *data)
{
	amilom7440_set_radio(buffer[0] == '0' ? AMILOM7440_RADIO_OFF : AMILOM7440_RADIO_ON);
	
	return count;
}

static int proc_get_radio(char *page, char **start, off_t offset,
			  int count, int *eof, void *data)
{
	int len = 0;
	
	len += snprintf(page, count, DRV_NAME ": %d\n", 
			amilom7440_radio_status == AMILOM7440_RADIO_OFF ? 0 : 1);
	
	*eof = 1;
	return len;
}


static void amilom7440_proc_cleanup(void)
{
	if (dir_base) {
		remove_proc_entry("radio", dir_base);
		remove_proc_entry(DRV_NAME, &proc_root);
		dir_base = NULL;
	}
}


static int amilom7440_proc_init(void)
{
	struct proc_dir_entry *ent;
	int err = 0;

	dir_base = create_proc_entry(DRV_NAME, S_IFDIR, &proc_root);
	if (dir_base == NULL) {
		printk(KERN_ERR DRV_NAME ": Unable to initialise /proc/" 
		       DRV_NAME "\n");
		err = -ENOMEM;
		goto fail;
	}


	ent = create_proc_entry("radio", S_IFREG | S_IRUGO | S_IWUSR, 
				dir_base);
	if (ent) {
		ent->read_proc = proc_get_radio;
		ent->write_proc = proc_set_radio;
	} else {
		printk(KERN_ERR
		       "Unable to initialize /proc/" DRV_NAME "/radio\n");
		err = -ENOMEM;
		goto fail;
	}

	return 0;

 fail:
	amilom7440_proc_cleanup();
	return err;
}

/*
 * module stuff
 */
static int __init amilom7440_init(void)
{
	amilom7440_proc_init();

	amilom7440_set_radio((radio == 1) ? AMILOM7440_RADIO_ON : AMILOM7440_RADIO_OFF);
	
	return 0;
}

static void __exit amilom7440_exit(void)
{
	amilom7440_set_radio(AMILOM7440_RADIO_OFF);

	amilom7440_proc_cleanup();
}

module_init(amilom7440_init);
module_exit(amilom7440_exit);

